# cython: language_level=3

# Python imports
from zoneinfo import available_timezones as _available_timezones

# Constants -----------------------------------------------------------------------------------
AVAILABLE_TIMEZONES: set[str] = _available_timezones()
del _available_timezones


########## All rest utility functions are in the utils.pxd file ##########
########## The following functions are for testing purpose only ##########


def _test_utils() -> None:
    _test_encode_decode_utf8()
    _test_encode_decode_ascii()


def _test_encode_decode_utf8() -> None:
    val = "中国\n한국어\nにほんご\nEspañol"
    # encode
    n = val.encode("utf-8")
    x = encode_str(val, "utf-8")  # type: ignore
    assert n == x, f"{n} | {x}"
    # decode
    i = n.decode("utf-8")
    j = decode_bytes(n, "utf-8")  # type: ignore
    k = decode_bytes_utf8(n)  # type: ignore
    assert i == j == k == val, f"{i} | {j} | {k} | {val}"
    print("Pass Encode/Decode UTF-8".ljust(80))


def _test_encode_decode_ascii() -> None:
    val = "hello\nworld"
    # encode
    n = val.encode("ascii")
    x = encode_str(val, "ascii")  # type: ignore
    assert n == x, f"{n} | {x}"
    # decode
    i = n.decode("ascii")
    j = decode_bytes(n, "ascii")  # type: ignore
    k = decode_bytes_ascii(n)  # type: ignore
    assert i == j == k == val, f"{i} | {j} | {k} | {val}"
    print("Pass Encode/Decode ASCII".ljust(80))



# cython: language_level=3

from cpython.bytes cimport (
    PyBytes_Size as bytes_len,
    PyBytes_AsString as bytes_to_chars,
)
from cpython.unicode cimport (
    PyUnicode_Decode,
    PyUnicode_DecodeUTF8,
    PyUnicode_DecodeASCII,
    PyUnicode_AsEncodedString,
    PyUnicode_GET_LENGTH as str_len,
    PyUnicode_Substring as str_substr,
)
cimport numpy as np
from numpy cimport (
    NPY_ORDER,
    PyArray_Flatten,
    PyArray_GETPTR1,
    PyArray_GETPTR2,
    PyArray_GETPTR3,
    PyArray_GETPTR4,
    PyArray_GETITEM,
    PyArray_SETITEM,
)

# Constants
cdef:
    set AVAILABLE_TIMEZONES

# Utils: encode / decode
cdef inline bytes encode_str(object obj, char* encoding):
    """Encode string to bytes using the 'encoding' with
    'surrogateescape' error handling `<'bytes'>`."""
    return PyUnicode_AsEncodedString(obj, encoding, b"surrogateescape")

cdef inline str decode_bytes(object data, char* encoding):
    """Decode bytes to string using the 'encoding' with
    "surrogateescape" error handling `<'str'>`."""
    return PyUnicode_Decode(bytes_to_chars(data), bytes_len(data), encoding, b"surrogateescape")

cdef inline str decode_bytes_utf8(object data):
    """Decode bytes to string using 'utf-8' encoding with
    'surrogateescape' error handling `<'str'>`."""
    return PyUnicode_DecodeUTF8(bytes_to_chars(data), bytes_len(data), b"surrogateescape")

cdef inline str decode_bytes_ascii(object data):
    """Decode bytes to string using 'ascii' encoding with
    'surrogateescape' error handling `<'str'>`."""
    return PyUnicode_DecodeASCII(bytes_to_chars(data), bytes_len(data), b"surrogateescape")

# NumPy: ndarray get item
cdef inline object arr_getitem_1d(np.ndarray arr, np.npy_intp i):
    """Get item from 1-dimensional numpy ndarray as `<'object'>`."""
    cdef void* itemptr = <void*>PyArray_GETPTR1(arr, i)
    return PyArray_GETITEM(arr, itemptr)

cdef inline object arr_getitem_2d(np.ndarray arr, np.npy_intp i, np.npy_intp j):
    """Get item from 2-dimensional numpy ndarray as `<'object'>`."""
    cdef void* itemptr = <void*>PyArray_GETPTR2(arr, i, j)
    return PyArray_GETITEM(arr, itemptr)

cdef inline object arr_getitem_3d(np.ndarray arr, np.npy_intp i, np.npy_intp j, np.npy_intp k):
    """Get item from 3-dimensional numpy ndarray as `<'object'>`."""
    cdef void* itemptr = <void*>PyArray_GETPTR3(arr, i, j, k)
    return PyArray_GETITEM(arr, itemptr)

cdef inline object arr_getitem_4d(np.ndarray arr, np.npy_intp i, np.npy_intp j, np.npy_intp k, np.npy_intp l):
    """Get item from 4-dimensional numpy ndarray as `<'object'>`."""
    cdef void* itemptr = <void*>PyArray_GETPTR4(arr, i, j, k, l)
    return PyArray_GETITEM(arr, itemptr)

# NumPy: ndarray set item
cdef inline bint arr_setitem_1d(np.ndarray arr, object item, np.npy_intp i) except -1:
    """Set item for 1-dimensional numpy ndarray."""
    cdef void* itemptr = <void*>PyArray_GETPTR1(arr, i)
    return PyArray_SETITEM(arr, itemptr, item)

cdef inline bint arr_setitem_2d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j) except -1:
    """Set item for 2-dimensional numpy ndarray."""
    cdef void* itemptr = <void*>PyArray_GETPTR2(arr, i, j)
    return PyArray_SETITEM(arr, itemptr, item)

cdef inline bint arr_setitem_3d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j, np.npy_intp k) except -1:
    """Set item for 3-dimensional numpy ndarray."""
    cdef void* itemptr = <void*>PyArray_GETPTR3(arr, i, j, k)
    return PyArray_SETITEM(arr, itemptr, item)

cdef inline bint arr_setitem_4d(np.ndarray arr, object item, np.npy_intp i, np.npy_intp j, np.npy_intp k, np.npy_intp l) except -1:
    """Set item for 4-dimensional numpy ndarray."""
    cdef void* itemptr = <void*>PyArray_GETPTR4(arr, i, j, k, l)
    return PyArray_SETITEM(arr, itemptr, item)

# NumPy: ndarray flatten
cdef inline np.ndarray arr_flatten(np.ndarray arr):
    """Flatten multi-dimensional numpy ndarray into 
    1-dimension with 'C-ORDER' `<'np.ndarray'>`."""
    return PyArray_Flatten(arr, NPY_ORDER.NPY_CORDER)

# NumPy: nptime unit
cdef inline str map_nptime_unit_int2str(int unit):
    """Map ndarray[datetime64/timedelta64] time unit from integer
    to the corresponding string representation `<'str'>`."""
    # Common units
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_ns:
        return "ns"
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_us:
        return "us"
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_ms:
        return "ms"
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_s:
        return "s"
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_m:
        return "m"
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_h:
        return "h"
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_D:
        return "D"
    # Uncommon units
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_Y:
        return "Y"
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_M:
        return "M"
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_W:
        return "W"
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_ps:
        return "ps"
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_fs:
        return "fs"
    if unit == np.NPY_DATETIMEUNIT.NPY_FR_as:
        return "as"
    # if unit == np.NPY_DATETIMEUNIT.NPY_FR_B:
    #     return "B"
    raise ValueError("Unsupported numpy time unit: %d." % (unit))

cdef inline int map_nptime_unit_str2int(str unit):
    """Map ndarray[datetime64/timedelta64] time unit from string
    representation to the corresponding integer `<'int'>`."""
    # Common units
    if unit == "ns":
        return np.NPY_DATETIMEUNIT.NPY_FR_ns
    if unit == "us":
        return np.NPY_DATETIMEUNIT.NPY_FR_us
    if unit == "ms":
        return np.NPY_DATETIMEUNIT.NPY_FR_ms
    if unit == "s":
        return np.NPY_DATETIMEUNIT.NPY_FR_s
    if unit == "m":
        return np.NPY_DATETIMEUNIT.NPY_FR_m
    if unit == "h":
        return np.NPY_DATETIMEUNIT.NPY_FR_h
    if unit == "D":
        return np.NPY_DATETIMEUNIT.NPY_FR_D
    # Uncommon units
    if unit == "Y":
        return np.NPY_DATETIMEUNIT.NPY_FR_Y
    if unit == "M":
        return np.NPY_DATETIMEUNIT.NPY_FR_M
    if unit == "W":
        return np.NPY_DATETIMEUNIT.NPY_FR_W
    if unit == "ps":
        return np.NPY_DATETIMEUNIT.NPY_FR_ps
    if unit == "fs":
        return np.NPY_DATETIMEUNIT.NPY_FR_fs
    if unit == "as":
        return np.NPY_DATETIMEUNIT.NPY_FR_as
    # if unit == "B":
    #     return np.NPY_DATETIMEUNIT.NPY_FR_B
    raise ValueError("Unsupported numpy time unit: %s." % (unit))

cdef inline int parse_arr_nptime_unit(np.ndarray arr):
    """Parse numpy datetime64/timedelta64 time unit from the
    given 'arr', returns the unit in `<'int'>`."""
    cdef:
        str dtype_str = arr.dtype.str
        Py_ssize_t length = str_len(dtype_str)
    if length < 6:
        raise ValueError("Failed to parse arr time unit from: '%s'." % dtype_str)
    return map_nptime_unit_str2int(str_substr(dtype_str, 4, length - 1))



# cython: language_level=3
# cython: wraparound=False
# cython: boundscheck=False

# Cython imports
import cython
from cython.cimports.serializor.binary.ser import serialize as _ser_bin  # type: ignore
from cython.cimports.serializor.binary.des import deserialize as _des_bin  # type: ignore

# Python imports
from hashlib import sha256 as _sha256
from base64 import urlsafe_b64encode as _urlsafe_b64encode
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.backends import default_backend as _default_backend
from serializor.binary.ser import serialize as _ser_bin
from serializor.binary.des import deserialize as _des_bin
from serializor import errors


# Utils -----------------------------------------------------------------------------
@cython.cfunc
@cython.inline(True)
def _generate_salt(key: bytes) -> bytes:
    """(cfunc) Generate a salt form the given key `<'bytes'>`."""
    array: bytearray = bytearray()
    byte: cython.uint
    shift: cython.uint = 1
    for byte in key:
        array.append((byte + shift) % 256)
        shift = 1 if shift > 10 else shift + 1
    return _sha256(array[::-1]).digest()


@cython.cfunc
@cython.inline(True)
def _create_fernet(key: bytes) -> object:
    """(cfunc) Create a Fernet from the given key `<'Fernet'>`."""
    return Fernet(
        _urlsafe_b64encode(
            PBKDF2HMAC(
                algorithm=SHA256(),
                length=32,
                salt=_generate_salt(key),
                iterations=32,
                backend=_default_backend(),
            ).derive(key)
        )
    )


# Encrypt ----------------------------------------------------------------------------
@cython.ccall
def encrypt(obj: object, key: object) -> bytes:
    """Serialize and encrypt an object with the given 'key' into `<'bytes'>`.

    :param obj `<'object'>`: The object to encrypt.
    :param key `<'object'>`: The key to encrypt the object with.
    """
    # Serialize & Encrypt
    val: bytes = _ser_bin(obj)
    fernet: Fernet = _create_fernet(_ser_bin(key))
    try:
        return fernet.encrypt(val)
    except Exception as err:
        raise errors.EncryptError(
            "<'Serializor'>\nObject encryption failed: %s" % err
        ) from err


# Decrypt ----------------------------------------------------------------------------
@cython.ccall
def decrypt(data: bytes, key: object) -> object:
    """Decrypt and deserialize encrypted data (bytes)
    with the given 'key' back to an `<'object'>`.

    :param enc `<'bytes'>`: The encrypted data.
    :param key `<'object'>`: The key to decrypt the data with.
    """
    # Decrypt & Deserialize
    fernet: Fernet = _create_fernet(_ser_bin(key))
    try:
        val = fernet.decrypt(data)
    except Exception as err:
        raise errors.DecryptError(
            "<'Serializor'>\nData decryption failed: %s" % err
        ) from err
    return _des_bin(val)



# cython: language_level=3

cdef:
    # Basic Types
    char STR
    char BOOL
    char FLOAT
    char INT
    char NONE

    # Date&Time Types
    char DATE
    char TIME
    char DATETIME
    char TIMEDELTA
    char STRUCT_TIME

    # Numeric Types
    char DECIMAL
    char COMPLEX

    # Bytes Types
    char BYTES
    char BYTEARRAY

    # Sequence Types
    char LIST
    char TUPLE
    char SET
    char FROZENSET
    char RANGE

    # Mapping Types
    char DICT

    # NumPy Types
    char DATETIME64
    char TIMEDELTA64
    char NDARRAY
    char NDARRAY_OBJECT
    char NDARRAY_INT
    char NDARRAY_UINT
    char NDARRAY_FLOAT
    char NDARRAY_BOOL
    char NDARRAY_DT64
    char NDARRAY_TD64
    char NDARRAY_COMPLEX
    char NDARRAY_BYTES
    char NDARRAY_UNICODE

    # Pandas Types
    char SERIES
    char DATAFRAME
    char PD_TIMESTAMP
    char PD_TIMEDELTA
    char DATETIMEINDEX
    char TIMEDELTAINDEX

    ### Test for duplicates ###
    bint _duplicates_test



# cython: language_level=3
# cython: wraparound=False
# cython: boundscheck=False

# Cython imports
import cython
from cython.cimports.serializor.binary.ser import serialize as _ser_bin  # type: ignore
from cython.cimports.serializor.unicode.ser import serialize as _ser_uni  # type: ignore

# Python imports
from serializor.binary.ser import serialize as _ser_bin
from serializor.unicode.ser import serialize as _ser_uni

__all__ = ["serialize"]


# Serialize -----------------------------------------------------------------------------------
@cython.ccall
def serialize(obj: object, to_bytes: cython.bint = True) -> object:
    """Serialize an object to `<'bytes'>` or `<'str'>`.

    :param obj `<'object'>`: The object to serialize.
    :param to_bytes `<'bool'>`: Whether to serialize to `<'bytes'>` (recommended). Defaults to `True`.

    Supports:
    - Python natives:
        - Ser <'str'> -> Des <'str'>
        - Ser <'int'> -> Des <'int'>
        - Ser <'float'> -> Des <'float'>
        - Ser <'bool'> -> Des <'bool'>
        - Ser <'datetime.datetime'> -> Des <'datetime.datetime'> `[Supports Timezone]`
        - Ser <'datetime.date'> -> Des <'datetime.date'>
        - Ser <'datetime.time'> -> Des <'datetime.time'> `[Supports Timezone]`
        - Ser <'datetime.timedelta'> -> Des <'datetime.timedelta'>
        - Ser <'time.struct_time'> -> Des <'time.struct_time'>
        - Ser <'decimal.Decimal'> -> Des <'decimal.Decimal'>
        - Ser <'complex'> -> Des <'comples'>
        - Ser <'bytes'> -> Des <'bytes'>
        - Ser <'bytearray'> -> Des <'bytearray'>
        - Ser <'memoryview'> -> Des <'bytes'>
        - Ser <'list'> -> Des <'list'>
        - Ser <'tuple'> -> Des <'tuple'>
        - Ser <'set'> -> Des <'set'>
        - Ser <'frozenset'> -> Des <'frozenset'>
        - Ser <'range'> -> Des <'range'>
        - Ser <'deque'> -> Des <'list'>
        - Ser <'dict'> -> Des <'dict'>
    - NumPy objects:
        - Ser <'np.str\_'> -> Des <'str'>
        - Ser <'np.int\*'> -> Des <'int'>
        - Ser <'np.uint\*'> -> Des <'int'>
        - Ser <'np.float\*'> -> Des <'float'>
        - Ser <'np.bool\_'> -> Des <'bool'>
        - Ser <'np.datetime64'> -> Des <'np.datetime64'>
        - Ser <'np.timedelta64'> -> Des <'np.timedelta64'>
        - Ser <'np.complex\*'> -> Des <'complex'>
        - Ser <'np.bytes\_'> -> Des <'bytes'>
        - Ser <'np.ndarray'> -> Des <'np.ndarray'> `[1-4 dimemsional]`
    - Pandas objects:
        - Ser <'pd.Timestamp'> -> Des <'pd.Timestamp'> `[Supports Timezone]`
        - Ser <'pd.Timedelta'> -> Des <'pd.Timedelta'>
        - Ser <'pd.Series'> -> Des <'pd.Series'>
        - Ser <'pd.DatetimeIndex'> -> Des <'pd.DatetimeIndex'> `[Supports Timezone]`
        - Ser <'pd.TimedeltaIndex'> -> Des <'pd.TimedeltaIndex'>
        - Ser <'pd.DataFrame'> -> Des <'pd.DataFrame'>
    """
    return _ser_bin(obj) if to_bytes else _ser_uni(obj)



# /usr/bin/python
# -*- coding: UTF-8 -*-
from serializor.ser import serialize
from serializor.des import deserialize
from serializor.crypto import encrypt, decrypt
from serializor.errors import (
    SerializorError,
    SerializeError,
    DeserializeError,
    CryptoError,
    EncryptError,
    DecryptError,
)

__all__ = [
    # functions
    "serialize",
    "deserialize",
    "encrypt",
    "decrypt",
    # errors
    "SerializorError",
    "SerializeError",
    "DeserializeError",
    "CryptoError",
    "EncryptError",
    "DecryptError",
]
